// 制約クラスで IUIComponentConstraint を実装します。
// エレメントの制約に関する設定は IUIComponentConstraint を実装したクラスで設定できます。
class Constraint implements IUIComponentConstraint {

  // エレメントの移動可否を指定するメソッドです。
  // true を返却する場合、デザイナ画面上に配置したエレメントを移動できます。
  // このサンプルでは、エレメントは移動可能にするため、true を返却します。
  public get movable(): boolean {
    return true;
  }

  // エレメントの削除可否を指定するメソッドです。
  // true を返却する場合、デザイナ画面上に配置したエレメントを削除できます。
  // このサンプルでは、エレメントは削除可能にするため、true を返却します。
  public get removable(): boolean {
    return true;
  }

  // エレメントの複製可否を指定するメソッドです。
  // true を返却する場合、デザイナ画面上に配置したエレメントを複製できます。
  // このサンプルでは、子エレメントを作成しませんので、true を返却します。
  public get copyable(): boolean {
    return true;
  }

  // エレメントが末端かどうかを指定するメソッドです。
  // true を返却する場合、デザイナ画面上に配置したエレメントの配下に子エレメントは配置できません。
  // このメソッドの返却値に true を指定した場合は、acceptableChild() の返却値は常に false とします。
  // このサンプルでは、子エレメントを作成しませんので、true を返却します。
  public get extremity(): boolean {
    return true;
  }

  // 親要素に対して、自身の配置可否を指定するメソッドです。
  // true を返却した場合、デザイナ画面上に配置したエレメントを親として、自身を子エレメントに配置できます。
  // parent 引数には親エレメントが渡されますので、親エレメントによって配置可能とするかどうか指定できます。
  // このサンプルでは、どの親エレメントに対しても配置可能としますので、常に true を返却します。
  public acceptableParent(parent: ParentUIElement): boolean {
    return true;
  }

  // 自身に対して、子エレメントの配置可否を指定するメソッドです。
  // true を返却した場合、デザイナ画面上に配置した自身のエレメントに対して、子エレメントを配置できます。
  // child 引数には子エレメントが渡されますので、子エレメントによって配置可能とするかどうか指定できます。
  // 常に false を返却する場合は、extremity() の返却値は常に true とします。
  // このサンプルでは、自身より子階層にエレメントを移動させたくないため、常に false を返却します。
  public acceptableChild(child: IUIElement): boolean {
    return false;
  }
}

// エレメント固有カテゴリのプロパティ一覧を設定できます。
type PropertyDefinition = {
  value: UniquePropertyDefinitionType;
};

// エレメント固有カテゴリのプロパティは uniquePropertyDefinition で設定できます。
const uniquePropertyDefinition: IUniquePropertyDefinition & PropertyDefinition = {
  value: {
    displayName: 'value',
    definition: {
      // value プロパティには何らかの指定が必要である
      required: true,
      // valueプロパティの値が変わった時に再レンダリングされる必要がある
      rerender: true,
    },
    type: 'string',
  },
};

export class MyZipCodeField implements IUIElementCore {
  // 子要素のビルダは updateElement メソッド内からもアクセスされるため、クラスの先頭で宣言しておきます。
  private _firstInputBuilder: IHTMLElementBuilder;
  private _secondInputBuilder: IHTMLElementBuilder;

  // 繰り返しのないエレメントなので、SimpleUIElement を指定します。
  public get wrapperClass(): UIElementWrapperClass {
    return 'SimpleUIElement';
  }

  // エレメント名を返却するメソッドです。
  // { エレメントのクラス名 } .name 形式で、エレメントのクラス名を返却します。
  // ここでは、MyZipCodeField という名前にします。
  public get elementTypeName(): string {
    return MyZipCodeField.name;
  }

  // エレメントの表示形式を返却します。
  public get displayType(): UIElementDisplayType {
    return 'INPUT';
  }

  // 上で実装した制約クラスのクラス名 Constraint を返却します。
  public get constraintClass(): UIComponentConstraintClass {
    return Constraint;
  }

  // エレメント固有プロパティの定義を返却します。
  public get uniquePropertyDefinition(): IUniquePropertyDefinition {
    return uniquePropertyDefinition;
  }

  // 共通プロパティの定義を返却します。
  public get commonPropertyDefinition(): ICommonPropertyDefinition {
    return {};
  }

  // createChildren(), updateChildren(), cloneChildren() で
  // 子エレメントとして自身以外のエレメントを使用する場合に、そのクラス名を返却します。
  // ここで返却されたエレメントのクラスは、IM-BloomMaker 側によって自動的にクラスが登録されるため、
  // 上記の3メソッドの中で使用できます。
  // このサンプルでは子エレメントを作成しませんので、空配列を返却します。
  public get dependElements(): UIElementCoreClass[] {
    return [];
  }

  // エレメントが作成された際に一度だけ呼び出されるメソッドです。
  // このメソッド内で、このエレメントで使用するルートとなる HTML エレメントを、HTMLElementBuilder を経由して返却します。
  // updateElement() によって都度変更されない常時適用する属性値やイベントの設定は、このメソッドに実装します。
  // このサンプルでは、子要素の作成、イベントリスナ、属性、スタイルの付与を行っています。
  public createElement(
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<PropertyDefinition & ICommonPropertyDefinition>
  ): IHTMLElementBuilder {
    // エレメントを作成します。
    const builder = window.imHichee.HTMLElementBuilder.createElement('div', HTMLDivElement);

    // 子タグのテキストボックスを作成します。
    const inputElement1 = document.createElement('input');
    const inputElement2 = document.createElement('input');

    // テキストボックスの input イベントが呼ばれると、valueプロパティを更新するようにリスナを付与します。
    inputElement1.addEventListener('input', (e) => {
        properties.setProperty('value', inputElement1.value + inputElement2.value, false);
    });
    inputElement2.addEventListener('input', (e) => {
        properties.setProperty('value', inputElement1.value + inputElement2.value, false);
    });

    this._firstInputBuilder = window.imHichee.HTMLElementBuilder.fromElement(inputElement1);
    this._secondInputBuilder = window.imHichee.HTMLElementBuilder.fromElement(inputElement2);
    const labelElement = window.imHichee.HTMLElementBuilder.createElement('label', HTMLLabelElement).setText('-');

    // setAttribute() を使用すると、作成したエレメントに属性を追加できます。
    builder.setAttribute('my-attribute', 'zip-code-input');

    // setCSS() を使用すると、作成したエレメントにstyleを追加できます。
    builder.setCSS('padding', '10px');

    // appendChild() を利用して、上で作成したエレメント（ビルダ）を子エレメントとして追加します。
    // ここで追加した子エレメントは、IM-BloomMakerのデザイナ上で1つ1つの要素を動かしたりプロパティを設定したりすることはできません。全体で1つのエレメントとして扱われます。
    builder
        .appendChild(this._firstInputBuilder)
        .appendChild(labelElement)
        .appendChild(this._secondInputBuilder);

    // 返却値には作成したエレメントを指定します。
    return builder;
  }

  // レンダリングが必要になった際に呼び出されるメソッドです。
  // プロパティで紐づけている変数の値が変更された際など、エレメント側を更新する必要がある場合は、このメソッドに実装します。
  // このサンプルでは、プロパティである value が変更された際に、子要素の <input> タグに value の値を反映させる処理を記述していきます。
  public updateElement(
    builder: IHTMLElementBuilder,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<PropertyDefinition & ICommonPropertyDefinition>
  ): void {
    // プロパティの値を文字列で取得します。
    const value = properties.getProperty('value').toString();

    // 取得したプロパティの値を2つの入力エレメントに反映するため、前3桁と残りに分割します。
    const value1 = value.substring(0, 3);
    const value2 = value.substring(3);

    // 子ノードの<input>タグのvalue属性に値をセットします。
    // 子ノードの IHTMLElementBuilder は自動で build() が呼ばれない仕様であるため、独自で build() を実行する必要があります。
    this._firstInputBuilder.setValue(value1).build();
    this._secondInputBuilder.setValue(value2).build();
  }

  // レンダリングが行われた後に、レンダリングが不要になった際に呼び出されるメソッドです。
  // document.body や window などグローバルに対してのイベントの設定を解除する処理を、このメソッドに実装します。
  // 特に実装がない場合は、このメソッドの実装は不要です。
  // このサンプルでは、便宜上実装していますが、何も処理がなければ実装を削除できます。
  public hiddenElement(
    builder: IHTMLElementBuilder,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<PropertyDefinition & ICommonPropertyDefinition>
  ): void {
    return;
  }

  // エレメントが削除された際に呼び出されるメソッドです。
  // 外部ライブラリを使用しており、ライブラリの解放処理が必要な場合に、このメソッドに実装します。
  // そのため、エレメントの実装のみで表示処理が完結している場合は、このメソッドの実装は不要です。
  // このサンプルでは、便宜上実装していますが、何も処理がなければ実装を削除できます。
  public destroyElement(
    builder: IHTMLElementBuilder,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): void {
    return;
  }

  // エレメントが作成された際に一度だけ呼び出されるメソッドです。 createElement() の後に呼び出されます。
  // 自身のエレメント配下に子エレメントを作成する必要がある場合、このメソッドでエレメントを作成し、配列として返却します。
  // 子エレメントが不要な場合は、空配列を返却します。
  // IHTMLElementBuilder.appendChild() で子要素を追加する場合と異なり、
  // デザイナ上でコンテナに配置した際に、子要素はそれぞれ別のエレメントとして扱われます。
  public createChildren(self: IUIElement, container: IUIContainer): IUIElement[] {
    return [];
  }

  // レンダリングが必要になった際に呼び出されるメソッドです。 updateElement() の後に呼び出されます。
  // プロパティの状態によって子エレメントの数に変更がある場合、新しい子エレメントの配列を返却することで、子エレメントを再配置します。
  // 子エレメントの数に変更がない場合でも、子エレメントのプロパティ値を変更する場合は、このメソッドに実装します。
  // createChildren() で空配列を返却しており、子エレメントを追加する予定がない場合は、このメソッドは実装しません。
  // このサンプルでは、便宜上実装していますが、何も処理がなければ実装を削除できます。
  public updateChildren(
    children: IUIElement[],
    self: IUIElement,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IUIElement[] {
    return children;
  }

  // エレメントが複製された際に呼び出されるメソッドです。 createElement() の後に呼び出されます。
  // children 引数には、同様に複製されたエレメントが格納されています。
  // プロパティの状態によって子エレメントの数に変更がある場合、新しい子エレメントの配列を返却することで、子エレメントを再配置します。
  // 子エレメントの数に変更がない場合でも、子エレメントのプロパティ値を変更する場合は、このメソッドに実装します。
  // 子エレメントに対して何も操作しない場合は、children 引数をそのまま返却します。
  public clonedChildren(
    children: IUIElement[],
    self: IUIElement,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IUIElement[] {
    return children;
  }
}
